/* Auto generated from checkpoint checkpoint_epoch_1.pth */


#ifndef DRED_RDOVAE_STATS_DATA_H
#define DRED_RDOVAE_STATS_DATA_H

#include "nnet.h"


#include "opus_types.h"

#include "dred_rdovae_constants.h"


extern const opus_uint8 dred_latent_quant_scales_q8[400];
extern const opus_uint8 dred_latent_dead_zone_q8[400];
extern const opus_uint8 dred_latent_r_q8[400];
extern const opus_uint8 dred_latent_p0_q8[400];


extern const opus_uint8 dred_state_quant_scales_q8[800];
extern const opus_uint8 dred_state_dead_zone_q8[800];
extern const opus_uint8 dred_state_r_q8[800];
extern const opus_uint8 dred_state_p0_q8[800];


#endif /* DRED_RDOVAE_STATS_DATA_H */
